{*******************************************************}
{File:      NCOciBuff.PAS                               }
{Revision:  0.02.00 / 28.10.1999                        }
{Comment:   NC OCI8 VCL: scrollable cursor implement.   }
{Copyright: (c) 1999, Nica-Com Ltd.                     }
{Author:    Dmitry Arefiev, diman@ncom.ru               }
{*******************************************************}
{$I NCOciDef.inc}
{*$DEFINE DEBUG}

unit NCOciBuff;

interface

Uses Classes, Forms, SysUtils, NCOciWrapper, NCOci;

type
    TOCICursor = class;
    TOCIDeadCursor = class;

    // ------------------------------------------------------------------------
    // Generic cursor
    // ------------------------------------------------------------------------

    TOCIRecordStatus = (rsEmpty, rsFetched, rsOld, rsNewInserted, rsNewUpdated,
        rsNewDeleted, rsNewInsApplyed, rsNewUpdApplyed, rsNewDelApplyed,
        rsDelphiBuff);
    POCIRecordStatus = ^TOCIRecordStatus;
    TOCIRecordStatuses = set of TOCIRecordStatus;
    TOCIUpdateAction = (uaFail, uaAbort, uaSkip, uaRetry, uaApplied);

    TOCICursorFilter = procedure (ASender: TOCICursor; var AAccept: Boolean) of object;
    TOCIUpdateRecordEvent = procedure (Sender: TOCIDeadCursor; var AAction: TOCIUpdateAction) of object;
    TOCIUpdateErrorEvent = procedure (Sender: TOCIDeadCursor; E: Exception; var AAction: TOCIUpdateAction) of object;

    POCIBookmark = pointer;

    TOCICursor = class
    private
        function GetDefVar(APosition: sb4): TOCIVariable;
        function GetDefVarCount: ub4;
    protected
        FDefVars: TList;
        FEOF, FBOF: Boolean;
        FOnFilterRecord: TOCICursorFilter;
        FCursorCanceled: Boolean;
        FCachedUpdates: Boolean;
        FInApplyUpdates: Boolean;
        procedure CheckActive;
        procedure CheckInactive;
        function GetRecNo: sb4; virtual; abstract;
        procedure SetRecNo(AValue: sb4); virtual; abstract;
        function GetActive: Boolean; virtual; abstract;
        function GetRecordCount: sb4; virtual; abstract;
        function GetRecordStatus: TOCIRecordStatus; virtual; abstract;
        function GetBookmark: POCIBookmark; virtual; abstract;
        procedure GotoBookmark(ABmk: POCIBookmark); virtual; abstract;
        function GetBookmarkSize: ub4; virtual; abstract;
        function GetRecordSize: ub4;
    public
        constructor Create;
        destructor Destroy; override;
        procedure AddDefVar(AVar: TOCIVariable);
        procedure RemoveDefVar(APosition: sb4);

        function GetData(ABmk: POCIBookmark; APosition: ub4; ABuff: pUb1;
            var ASize: sb4): Boolean; virtual; abstract;
        function GetDataPtr(ABmk: POCIBookmark; APosition: ub4; var ABuff: pUb1;
            var ASize: sb4): Boolean; virtual; abstract;
        procedure SetData(ABmk: POCIBookmark; APosition: ub4; ABuff: pUb1;
            ASize: sb4); virtual; abstract;
        procedure SetIsNull(ABmk: POCIBookmark; APosition: ub4;
            AIsNull: Boolean); virtual; abstract; // for LOB fields

        function AllocateRecord: POCIBookmark; virtual; abstract;
        procedure FreeRecord(P: POCIBookmark); virtual; abstract;
        procedure InitRecord(P: POCIBookmark); virtual; abstract;
        procedure GetRecord(P: POCIBookmark; PRef: POCIBookmark); virtual; abstract;
        procedure DeleteRecord; virtual; abstract;
        procedure ModifyRecord(P: POCIBookmark); virtual; abstract;
        procedure InsertRecord(P: POCIBookmark); virtual; abstract;
        function CompareBookmarks(P1, P2: POCIBookmark): sb2; virtual; abstract;
        function BookmarkValid(P: POCIBookmark): Boolean; virtual; abstract;

        procedure OpenComplete; virtual; abstract;
        procedure Close; virtual; abstract;
        procedure Skip(ANum: sb4); virtual; abstract;
        procedure First; virtual; abstract;
        procedure Last; virtual; abstract;

        function GetNewRecord(P: POCIBookmark): POCIBookmark; virtual; abstract;
        function GetOldRecord(P: POCIBookmark): POCIBookmark; virtual; abstract;
        procedure ApplyUpdates(P: POCIBookmark); virtual; abstract;
        procedure EndUpdate(P: POCIBookmark);

        property EOF: Boolean read FEOF;
        property BOF: Boolean read FBOF;
        property RecordSize: ub4 read GetRecordSize;
        property BookmarkSize: ub4 read GetBookmarkSize;
        property Bookmark: POCIBookmark read GetBookmark write GotoBookmark;
        property RecNo: sb4 read GetRecNo write SetRecNo;
        property RecordCount: sb4 read GetRecordCount;
        property RecordStatus: TOCIRecordStatus read GetRecordStatus;
        property Active: Boolean read GetActive;
        property DefVarCount: ub4 read GetDefVarCount;
        property DefVar[AIndex: sb4]: TOCIVariable read GetDefVar;

        property OnFilterRecord: TOCICursorFilter read FOnFilterRecord write FOnFilterRecord;
        property CursorCanceled: Boolean read FCursorCanceled;
        property CachedUpdates: Boolean read FCachedUpdates write FCachedUpdates;
        property InApplyUpdates: Boolean read FInApplyUpdates;
    end;

    // ------------------------------------------------------------------------
    // Scrollable cursor with dead caching
    // ------------------------------------------------------------------------

    POCIDeadBookmark = ^TOCIDeadBookmark;
    PPOCIDeadBookmark = ^POCIDeadBookmark;
    TOCIDeadBookmark = record
        FPageIndex: sb4;
        FRecIndex: sb4;
    end;

    POCIDeadRecordHeader = ^TOCIDeadRecordHeader;
    PPOCIDeadRecordHeader = ^POCIDeadRecordHeader; 
    TOCIDeadRecordHeader = record
        FSelfBmk: TOCIDeadBookmark;
        FStatus: TOCIRecordStatus;
        FRecordNumber: sb4;
        FRefRecord: POCIDeadRecordHeader;
    end;

    POCIDeadPage = pUb1;
    // TOCIDeadRecordHeader[0]
    // ...
    // TOCIDeadRecordHeader[FRecPerPage - 1]
    // DATA Var[1]
    // ...
    // DATA Var[FDefVars.Count]

    TUB4s = array[0..511] of ub4;
    PUB4s = ^TUB4s;

    TOCIDeadCursor = class(TOCICursor)
    private
        FStmt: TOCIStatement;
        FRecPerPage: sb4;
        FPageSize: sb4;
        FPages: TList;
        FStatusFilter: TOCIRecordStatuses;
        FPosition: TOCIDeadBookmark;
        FRowsInserted, FRowsDeleted: ub4;
        FVarOffsets: PUB4s;
        FFreeRecordsPage: sb4;
        FOnUpdateRecord: TOCIUpdateRecordEvent;
        FOnUpdateError: TOCIUpdateErrorEvent;
        FUnidirectional: Boolean;
        FOpenComplete: Boolean; 
        procedure PageUpdateSize;
        function PageNew: Integer;
        procedure PageFree(APageInd: ub4);
        function PageHeadersPtr(APageInd: ub4): POCIDeadRecordHeader;
        function PageDatasPtr(APageInd: ub4): pUb1;
        function PageRecordHeader(APageInd, ARecInd: ub4): POCIDeadRecordHeader;
        procedure PageBind(APageInd: ub4);
        procedure PageFetched(APageInd: ub4);
        procedure FreePages;
        function PageRecordData(APageInd: ub4; APosition: ub4): pUb1;
        procedure ValidatePosition(APageInd, ARecInd: sb4; AExcludeEO: Boolean);
        function P2RecHead(P: Pointer): POCIDeadRecordHeader;
        procedure CopyRecord(PCopyFrom, PCopyTo: POCIDeadBookmark);
        procedure ClearRecord(PRec: POCIDeadBookmark);
        function AllocateRecordInternal(AStatus: TOCIRecordStatus): POCIDeadRecordHeader;
        procedure SetCurrentRecord(P: POCIBookmark);
        procedure CheckModifable;
        procedure UpdateFreeRecPage(APageInd: ub4);
        procedure CheckBidir;
    protected
        function GetRecNo: sb4; override;
        procedure SetRecNo(AValue: sb4); override;
        function GetActive: Boolean; override;
        function GetRecordCount: sb4; override;
        function GetBookmark: POCIBookmark; override;
        procedure GotoBookmark(ABmk: POCIBookmark); override;
        function GetBookmarkSize: ub4; override;
        function GetRecordStatus: TOCIRecordStatus; override;
        function SetRecordStatus(APageInd, ARecInd: ub4; AStatus: TOCIRecordStatus): POCIDeadRecordHeader;
    public
        constructor Create;
        destructor Destroy; override;

        function GetData(ABmk: POCIBookmark; APosition: ub4; ABuff: pUb1;
            var ASize: sb4): Boolean; override;
        function GetDataPtr(ABmk: POCIBookmark; APosition: ub4; var ABuff: pUb1;
            var ASize: sb4): Boolean; override;
        procedure SetData(ABmk: POCIBookmark; APosition: ub4; ABuff: pUb1;
            ASize: sb4); override;
        procedure SetIsNull(ABmk: POCIBookmark; APosition: ub4;
            AIsNull: Boolean); override; // for LOB fields

        function AllocateRecord: POCIBookmark; override;
        procedure FreeRecord(P: POCIBookmark); override;
        procedure InitRecord(P: POCIBookmark); override;
        procedure GetRecord(P: POCIBookmark; PRef: POCIBookmark); override;
        procedure DeleteRecord; override;
        procedure ModifyRecord(P: POCIBookmark); override;
        procedure InsertRecord(P: POCIBookmark); override;
        function CompareBookmarks(P1, P2: POCIBookmark): sb2; override;
        function BookmarkValid(P: POCIBookmark): Boolean; override;

        procedure Open(AStmt: TOCIStatement; AService: TOCIService;
            AFetchSize: sb4; AExactRows: sb4; AUnidirectional: Boolean);
        procedure OpenComplete; override;
        procedure Describe(AStmt: TOCIStatement; AService: TOCIService);
        procedure Close; override;
        procedure Skip(ANum: sb4); override;
        procedure First; override;
        procedure Last; override;

        function GetNewRecord(P: POCIBookmark): POCIBookmark; override;
        function GetOldRecord(P: POCIBookmark): POCIBookmark; override;
        procedure ApplyUpdates(P: POCIBookmark); override;

        property StatusFilter: TOCIRecordStatuses read FStatusFilter write FStatusFilter;
        property OnUpdateRecord: TOCIUpdateRecordEvent read FOnUpdateRecord write FOnUpdateRecord;
        property OnUpdateError: TOCIUpdateErrorEvent read FOnUpdateError write FOnUpdateError;
    end;

implementation

Uses NCOciMsg;

// ------------------------------------------------------------------------
// Generic cursor
// ------------------------------------------------------------------------

constructor TOCICursor.Create;
begin
    inherited Create;
    FDefVars := TList.Create;
end;

destructor TOCICursor.Destroy;
begin
    FDefVars.Free;
    inherited Destroy;
end;

procedure TOCICursor.AddDefVar(AVar: TOCIVariable);
begin
    CheckInactive;
    while FDefVars.Count < AVar.Position do
        FDefVars.Add(nil);
    FDefVars[AVar.Position - 1] := AVar;
end;

procedure TOCICursor.CheckActive;
    procedure BuffMBOpen;
    begin
        OCIDBError(msgOCIBuffMBOpen);
    end;
begin
    if not Active then
        BuffMBOpen;
end;

procedure TOCICursor.CheckInactive;
    procedure BuffMBClose;
    begin
        OCIDBError(msgOCIBuffMBClose);
    end;
begin
    if Active then
        BuffMBClose;
end;

function TOCICursor.GetDefVarCount: ub4;
begin
    Result := FDefVars.Count;
end;

function TOCICursor.GetDefVar(APosition: sb4): TOCIVariable;
begin
    if (APosition < 1) or (APosition > FDefVars.Count) then
        OCIDBError(msgOCIInvalidVarIndex);
    Result := TOCIVariable(FDefVars[APosition - 1]);
end;

procedure TOCICursor.RemoveDefVar(APosition: sb4);
begin
    FDefVars.Delete(APosition - 1);
end;

function TOCICursor.GetRecordSize: ub4;
var
    i: Integer;
begin
    Result := 0;
    for i := 0 to FDefVars.Count - 1 do
        with TOCIVariable(FDefVars[i]) do
            Inc(Result, BindRecLen);
end;

procedure TOCICursor.EndUpdate(P: POCIBookmark);
begin
    if not CachedUpdates then
        ApplyUpdates(P);
end;

// ------------------------------------------------------------------------
// Scrollable cursor with dead caching
// ------------------------------------------------------------------------

constructor TOCIDeadCursor.Create;
begin
    inherited Create;
    FPages := TList.Create;
    FStatusFilter := [rsFetched, rsNewInserted, rsNewUpdated];
    FFreeRecordsPage := -1;
end;

destructor TOCIDeadCursor.Destroy;
begin
    if Active then
        Close;
    FPages.Free;
    inherited Destroy;
end;

function TOCIDeadCursor.GetActive: Boolean;
begin
    Result := FStmt <> nil;
end;

procedure TOCIDeadCursor.UpdateFreeRecPage(APageInd: ub4);
begin
    if (FRecPerPage > 1) and
       ((FFreeRecordsPage = -1) or (FFreeRecordsPage > APageInd)) then
        FFreeRecordsPage := APageInd;
end;

procedure TOCIDeadCursor.PageUpdateSize;
var
    i: Integer;
begin
    FPageSize := 0;
    GetMem(FVarOffsets, FDefVars.Count * SizeOf(Integer));
    for i := 0 to FDefVars.Count - 1 do
        with TOCIVariable(FDefVars[i]) do begin
            ArrayLen := FRecPerPage;
            FVarOffsets[i] := FPageSize;
            FPageSize := FPageSize + BindRecLen * FRecPerPage;
        end;
    FPageSize := FPageSize + ((SizeOf(TOCIDeadRecordHeader) + 3) and not 3) * FRecPerPage;
end;

function TOCIDeadCursor.SetRecordStatus(APageInd, ARecInd: ub4; AStatus: TOCIRecordStatus): POCIDeadRecordHeader;
const
    statWithHndl: TOCIRecordStatuses = [rsFetched, rsOld, rsNewInserted];

    procedure ForEachVar(AInit: Boolean; ANoHandleAlloc: Boolean);
    var
        i: Integer;
    begin
        for i := 0 to FDefVars.Count - 1 do
            with TOCIVariable(FDefVars[i]) do begin
                ArrayLen := FRecPerPage;
                ExternBuffer := PageRecordData(APageInd, i + 1);
                if AInit then
                    InitBuffer(ARecInd, ANoHandleAlloc)
                else
                    ClearBuffer(ARecInd, ANoHandleAlloc);
            end;
    end;

begin
    Result := PageRecordHeader(APageInd, ARecInd);
    with Result^ do begin
        if (FStatus in statWithHndl) and not (AStatus in statWithHndl) then
            ForEachVar(False, False)
        else if not (FStatus in statWithHndl) and (AStatus in statWithHndl) then
            ForEachVar(True, False);
        FStatus := AStatus;
    end;
end;

function TOCIDeadCursor.PageNew: Integer;
var
    i: ub4;
begin
    Result := FPages.Add(AllocMem(FPageSize));
    for i := 0 to FRecPerPage - 1 do
        SetRecordStatus(Result, i, rsEmpty);
end;

procedure TOCIDeadCursor.PageFree(APageInd: ub4);
var
    p: POCIDeadRecordHeader;
    i: ub4;
begin
    try
        for i := 0 to FRecPerPage - 1 do
            SetRecordStatus(APageInd, i, rsEmpty);
    finally
        p := PageHeadersPtr(APageInd);
        FreeMem(p, FPageSize);
        FPages.Delete(APageInd);
    end;
end;

procedure TOCIDeadCursor.FreePages;
begin
    while FPages.Count > 0 do
        PageFree(0);
    FreeMem(FVarOffsets);
end;

function TOCIDeadCursor.PageHeadersPtr(APageInd: ub4): POCIDeadRecordHeader;
begin
    Result := POCIDeadRecordHeader(FPages[APageInd]);
end;

function TOCIDeadCursor.PageRecordHeader(APageInd, ARecInd: ub4): POCIDeadRecordHeader;
begin
    ValidatePosition(APageInd, ARecInd, True);
    Result := POCIDeadRecordHeader(LongWord(PageHeadersPtr(APageInd)) +
        SizeOf(TOCIDeadRecordHeader) * ARecInd);
end;

function TOCIDeadCursor.PageDatasPtr(APageInd: ub4): pUb1;
begin
    Result := pUb1(PageRecordHeader(APageInd, FRecPerPage));
end;

function TOCIDeadCursor.PageRecordData(APageInd: ub4; APosition: ub4): pUb1;
begin
    Result := Pointer(LongWord(PageDatasPtr(APageInd)) + FVarOffsets[APosition - 1]);
end;

procedure TOCIDeadCursor.PageBind(APageInd: ub4);
var
    i: Integer;
begin
    for i := 0 to FRecPerPage - 1 do
        SetRecordStatus(APageInd, i, rsFetched);
    for i := 0 to FDefVars.Count - 1 do
        TOCIVariable(FDefVars[i]).Bind;
end;

procedure TOCIDeadCursor.PageFetched(APageInd: ub4);
var
    i: ub4;
    rNo: ub4;
begin
    FCursorCanceled := FStmt.EOF;
    if FStmt.LastRowCount = 0 then
        PageFree(APageInd)
    else begin
        rNo := RecordCount - FStmt.LastRowCount + 1;
        for i := 0 to FStmt.LastRowCount - 1 do
            with PageRecordHeader(APageInd, i)^ do begin
                FSelfBmk.FPageIndex := APageInd;
                FSelfBmk.FRecIndex := i;
                FRecordNumber := rNo + i;
                FRefRecord := nil;
            end;
        for i := FStmt.LastRowCount to FRecPerPage - 1 do
            SetRecordStatus(APageInd, i, rsEmpty);
        if FStmt.LastRowCount <= FRecPerPage - 1 then
            UpdateFreeRecPage(APageInd);
    end;
end;

procedure TOCIDeadCursor.Open(AStmt: TOCIStatement; AService: TOCIService;
    AFetchSize: sb4; AExactRows: sb4; AUnidirectional: Boolean);
var
    n: sb4;
begin
    CheckInactive;
    with FPosition do begin
        FPageIndex := -1;
        FRecIndex := AFetchSize - 1;
    end;
    FRecPerPage := AFetchSize;
    FCursorCanceled := False;
    FEOF := False;
    FBOF := False;
    FRowsInserted := 0;
    FRowsDeleted := 0;
    FStmt := AStmt;
    FUnidirectional := False;
    try
        try
            PageUpdateSize;
            PageNew;
            PageBind(0);
            if AExactRows > 0 then
                n := AExactRows
            else
                n := AFetchSize;
            FStmt.Execute(AService, n, AExactRows > 0, False, False);
        except
            FreePages;
            FStmt := nil;
            FCursorCanceled := True;
            raise;
        end;
        PageFetched(0);
        First;
    finally
        FUnidirectional := AUnidirectional;
    end;
    if not FUnidirectional and (AExactRows < 0) then begin
        Last;
        First;
    end;
end;

procedure TOCIDeadCursor.Describe(AStmt: TOCIStatement; AService: TOCIService);
begin
    FStmt := AStmt;
    FStmt.Describe(AService);
end;

procedure TOCIDeadCursor.Close;
begin
    CheckActive;
    FreePages;
    try
        if not FCursorCanceled then
            FStmt.CancelCursor;
    finally
        FStmt := nil;
    end;
end;

procedure TOCIDeadCursor.CheckBidir;
begin
    if FUnidirectional and FOpenComplete then
        OCIDBError(msgOCIMustBeBidir);
end;

procedure TOCIDeadCursor.OpenComplete;
begin
    FOpenComplete := True;
end;

procedure TOCIDeadCursor.Skip(ANum: sb4);
var
    d: Integer;
    actualNo: ub4;
    i: Integer;
    ok: Boolean;
begin
    CheckActive;
    if ANum < 0 then begin
        CheckBidir;
        actualNo := -ANum;
        d := -1;
    end
    else begin
        actualNo := ANum;
        d := 1;
    end;
    with FPosition do begin
        FBOF := (d = -1) and (FPageIndex = -1);
        FEOF := (d =  1) and (FPageIndex = FPages.Count) and FCursorCanceled;
        while (actualNo > 0) and not FEOF and not FBOF do begin
            FRecIndex := FRecIndex + d;
            if FRecIndex = -1 then begin
                FPageIndex := FPageIndex - 1;
                FRecIndex := FRecPerPage - 1;
                if FPageIndex = -1 then
                    FBOF := True;
            end
            else if FRecIndex = FRecPerPage then begin
                FPageIndex := FPageIndex + 1;
                FRecIndex := 0;
                if FPageIndex = FPages.Count then
                    if FCursorCanceled or (FStatusFilter = [rsEmpty]) then
                        FEOF := True
                    else begin
                        if not FUnidirectional then
                            i := PageNew
                        else
                            i := 0;
                        try
                            PageBind(i);
                            FStmt.Next(FRecPerPage);
                            PageFetched(i);
                        except
                            PageFree(i);
                            raise;
                        end;
                        if FCursorCanceled then
                            FEOF := FStmt.LastRowCount = 0;
                    end;
            end;
            if not FEOF and not FBOF and
               (PageRecordHeader(FPageIndex, FRecIndex)^.FStatus in FStatusFilter) then begin
                ok := True;
                if Assigned(FOnFilterRecord) then
                    FOnFilterRecord(Self, ok);
                if ok then
                    Dec(actualNo);
            end;
        end;
    end;
end;

procedure TOCIDeadCursor.First;
begin
    CheckActive;
    CheckBidir;
    with FPosition do begin
        FPageIndex := -1;
        FRecIndex := FRecPerPage - 1;
    end;
    Skip(1);
    FBOF := FEOF;
end;

procedure TOCIDeadCursor.Last;
begin
    CheckActive;
    CheckBidir;
    with FPosition do begin
        FPageIndex := FPages.Count - 1;
        FRecIndex := FRecPerPage - 1;
    end;
    Skip(MAXINT);
    Skip(-1);
    FEOF := FBOF;
end;

function TOCIDeadCursor.GetBookmarkSize: ub4;
begin
    Result := SizeOf(TOCIDeadBookmark);
end;

procedure TOCIDeadCursor.ValidatePosition(APageInd, ARecInd: sb4; AExcludeEO: Boolean);
    procedure InvBmk;
    begin
        OCIDBError(msgOCIInvalidBmk);
    end;
begin
    CheckActive;
    if AExcludeEO then begin
        if (APageInd < 0) or (APageInd > FPages.Count - 1) then
            InvBmk;
    end
    else begin
        if (APageInd < -1) or (APageInd > FPages.Count) then
            InvBmk;
    end;
    if (ARecInd < 0) or (ARecInd > FRecPerPage) then
        InvBmk;
end;

function TOCIDeadCursor.GetBookmark: POCIBookmark;
begin
    CheckActive;
    Result := @(PageRecordHeader(FPosition.FPageIndex, FPosition.FRecIndex)^.FSelfBmk);
end;

procedure TOCIDeadCursor.GotoBookmark(ABmk: POCIBookmark);
begin
    CheckActive;
    with POCIDeadBookmark(ABmk)^ do
        ValidatePosition(FPageIndex, FRecIndex, False);
    SetCurrentRecord(ABmk);
    if not FBOF and not FEOF then
        with FPosition, PageRecordHeader(FPageIndex, FRecIndex)^ do
            if (FRefRecord <> nil) and (FStatus = rsDelphiBuff) then
                FPosition := FRefRecord^.FSelfBmk;
end;

procedure TOCIDeadCursor.SetCurrentRecord(P: POCIBookmark);
begin
    FPosition := POCIDeadBookmark(P)^;
    FBOF := FPosition.FPageIndex = -1;
    FEOF := FPosition.FPageIndex = FPages.Count;
end;

function TOCIDeadCursor.GetRecNo: sb4;
begin
    CheckActive;
    if FEOF or FUnidirectional then
        Result := RecordCount
    else if FBOF then
        Result := 0
    else
        with FPosition do
            Result := PageRecordHeader(FPageIndex, FRecIndex)^.FRecordNumber;
end;

procedure TOCIDeadCursor.SetRecNo(AValue: sb4);
var
    pgNo, rNo: sb4;
begin
    CheckActive;
    CheckBidir;
    pgNo := (AValue - 1) div FRecPerPage;
    rNo := (AValue - 1) mod FRecPerPage;
    if FCursorCanceled and (pgNo >= FPages.Count) or (pgNo < 0) then
        OCIDBError(msgOCIInvalidRNO);
    if pgNo < FPages.Count then begin
        FPosition.FPageIndex := pgNo;
        FPosition.FRecIndex := rNo;
    end
    else begin
        FPosition.FPageIndex := FPages.Count - 1;
        FPosition.FRecIndex := FRecPerPage - 1;
    end;
    FEOF := False;
    FBOF := False;
    while not EOF and (RecNo < AValue) do
        Skip(1);
end;

function TOCIDeadCursor.GetRecordCount: sb4;
begin
    Result := FStmt.ROW_COUNT + FRowsInserted - FRowsDeleted;
end;

function TOCIDeadCursor.GetRecordStatus: TOCIRecordStatus;
begin
    CheckActive;
    if FEOF or FBOF then
        Result := rsEmpty
    else with FPosition do
        Result := PageRecordHeader(FPageIndex, FRecIndex)^.FStatus;
end;

function TOCIDeadCursor.GetData(ABmk: POCIBookmark; APosition: ub4;
    ABuff: pUb1; var ASize: sb4): Boolean;
begin
    if ABmk = nil then
        ABmk := @FPosition;
    with DefVar[APosition], POCIDeadBookmark(ABmk)^ do begin
        ExternBuffer := PageRecordData(FPageIndex, APosition);
        Result := GetData(FRecIndex, ABuff, ASize);
    end;
end;

function TOCIDeadCursor.GetDataPtr(ABmk: POCIBookmark; APosition: ub4;
    var ABuff: pUb1; var ASize: sb4): Boolean;
begin
    if ABmk = nil then
        ABmk := @FPosition;
    with DefVar[APosition], POCIDeadBookmark(ABmk)^ do begin
        ExternBuffer := PageRecordData(FPageIndex, APosition);
        Result := GetDataPtr(FRecIndex, ABuff, ASize);
    end;
end;

procedure TOCIDeadCursor.SetData(ABmk: POCIBookmark; APosition: ub4;
    ABuff: pUb1; ASize: sb4);
begin
    if ABmk = nil then
        ABmk := @FPosition;
    with DefVar[APosition], POCIDeadBookmark(ABmk)^ do begin
        ExternBuffer := PageRecordData(FPageIndex, APosition);
        SetData(FRecIndex, ABuff, ASize);
    end;
end;

procedure TOCIDeadCursor.SetIsNull(ABmk: POCIBookmark; APosition: ub4; AIsNull: Boolean);
begin
    if ABmk = nil then
        ABmk := @FPosition;
    with DefVar[APosition], POCIDeadBookmark(ABmk)^ do begin
        ExternBuffer := PageRecordData(FPageIndex, APosition);
        SetIsNull(FRecIndex, AIsNull);
    end;
end;

procedure TOCIDeadCursor.CopyRecord(PCopyFrom, PCopyTo: POCIDeadBookmark);
var
    i: Integer;
    p: pUb1;
    l: sb4;
begin
    for i := 1 to FDefVars.Count do begin
        GetDataPtr(PCopyFrom, i, p, l);
        SetData(PCopyTo, i, p, l);
    end;
end;

procedure TOCIDeadCursor.ClearRecord(PRec: POCIDeadBookmark);
var
    i: Integer;
begin
    for i := 1 to FDefVars.Count do
        SetData(PRec, i, nil, 0);
end;

// -----------------------------------------------------------------
// -----------------------------------------------------------------
// Delphi buffer management

function TOCIDeadCursor.P2RecHead(P: POCIBookmark): POCIDeadRecordHeader;
begin
    with POCIDeadBookmark(P)^ do
        Result := PageRecordHeader(FPageIndex, FRecIndex);
    if Result^.FStatus <> rsDelphiBuff then
        OCIDBError(msgOCIModifyNotDelphiBuff);
end;

function TOCIDeadCursor.AllocateRecordInternal(AStatus: TOCIRecordStatus): POCIDeadRecordHeader;
var
    savePos: TOCIDeadBookmark;
    saveStFlt: TOCIRecordStatuses;
begin
    CheckActive;
    savePos := FPosition;
    saveStFlt := FStatusFilter;
    try
        StatusFilter := [rsEmpty];
        // Free record page is unknown
        if FFreeRecordsPage = -1 then begin
            // If all records fetched then go to tail page
            // and try find free record
            if FCursorCanceled and (FPages.Count > 0) then begin
                FPosition.FPageIndex := FPages.Count - 1;
                FPosition.FRecIndex := 0;
                Skip(1);
            end
            // ... else allocate new empty page
            else
                FEOF := True;
        end
        // ... else try find empty record from FFreeRecordsPage page
        else begin
            FPosition.FPageIndex := FFreeRecordsPage - 1;
            FPosition.FRecIndex := FRecPerPage - 1;
            Skip(1);
        end;
        // Not found, then allocate page
        if FEOF then begin
            FPosition.FPageIndex := PageNew;
            FPosition.FRecIndex := 0;
        end;
        FFreeRecordsPage := FPosition.FPageIndex;
        Result := SetRecordStatus(FPosition.FPageIndex, FPosition.FRecIndex, AStatus);
        with Result^ do begin
            FSelfBmk := FPosition;
            FRecordNumber := FPosition.FPageIndex * FRecPerPage + FPosition.FRecIndex + 1;
            FRefRecord := nil;
        end;
    finally
        FPosition := savePos;
        FStatusFilter := saveStFlt;
    end;
end;

function TOCIDeadCursor.AllocateRecord: POCIBookmark;
begin
    Result := @(AllocateRecordInternal(rsDelphiBuff)^.FSelfBmk);
end;

procedure TOCIDeadCursor.FreeRecord(P: POCIBookmark);
begin
    P2RecHead(P)^.FRefRecord := nil; // ???
    with POCIDeadBookmark(p)^ do begin
        SetRecordStatus(FPageIndex, FRecIndex, rsEmpty);
        UpdateFreeRecPage(FPageIndex);
    end;
end;

procedure TOCIDeadCursor.InitRecord(P: POCIBookmark);
begin
    P2RecHead(P)^.FRefRecord := nil; // ???
    ClearRecord(P);
end;

procedure TOCIDeadCursor.GetRecord(P: POCIBookmark; PRef: POCIBookmark);
var
    pRH: POCIDeadRecordHeader;
begin
    pRH := PageRecordHeader(FPosition.FPageIndex, FPosition.FRecIndex);
    P2RecHead(P)^.FRefRecord := pRH;
    POCIDeadBookmark(PRef)^ := pRH^.FSelfBmk;
    CopyRecord(@FPosition, P);
end;

procedure TOCIDeadCursor.DeleteRecord;
var
    pCur: POCIDeadRecordHeader;
    pRH: POCIDeadRecordHeader;
    UndoValue: TOCIRecordStatus;
    pUndoVar: POCIRecordStatus;

    procedure InitUndo;
    begin
        pUndoVar := nil;
    end;

    procedure SaveUndo(AStatusVar: POCIRecordStatus);
    begin
        pUndoVar := AStatusVar;
        if pUndoVar <> nil then
            UndoValue := pUndoVar^;
    end;

    procedure DoUndo;
    begin
        if pUndoVar <> nil then
            pUndoVar^ := UndoValue;
    end;

begin
    CheckModifable;
    pRH := nil;
    pCur := PageRecordHeader(FPosition.FPageIndex, FPosition.FRecIndex);
    InitUndo;
    try
        with pCur^ do
            case FStatus of
            rsFetched, rsNewInserted, rsNewUpdated, rsNewDeleted:
                begin
                    SaveUndo(@FStatus);
                    with FPosition do
                        pRH := SetRecordStatus(FPageIndex, FRecIndex, rsNewDeleted);
                end;
            rsOld:
                begin
                    SaveUndo(@FRefRecord.FStatus);
                    with FRefRecord^.FSelfBmk do
                        pRH := SetRecordStatus(FPageIndex, FRecIndex, rsNewDeleted);
                end;
            rsEmpty, rsNewInsApplyed, rsNewUpdApplyed, rsNewDelApplyed, rsDelphiBuff:
                OCIDBError(msgOCICantModifyBuff);
            end;
        EndUpdate(@pRH.FSelfBmk);
    except
        DoUndo;
        raise;
    end;
    Inc(FRowsDeleted);
end;

procedure TOCIDeadCursor.ModifyRecord(P: POCIBookmark);
var
    pCur: POCIDeadRecordHeader;
    pRH: POCIDeadRecordHeader;
    pBmk: POCIDeadBookmark;
    UndoValue: TOCIRecordStatus;
    pUndoVar: POCIRecordStatus;
    UndoRefValue: POCIDeadRecordHeader;
    pUndoRefVar: PPOCIDeadRecordHeader;
    UndoFreeRec: POCIDeadRecordHeader;

    procedure InitUndo;
    begin
        pUndoVar := nil;
        pUndoRefVar := nil;
        UndoFreeRec := nil;
    end;

    procedure SaveUndo(AStatusVar: POCIRecordStatus;
        ARefRecord, AFreeRec: PPOCIDeadRecordHeader);
    begin
        pUndoVar := AStatusVar;
        if pUndoVar <> nil then
            UndoValue := pUndoVar^;
        pUndoRefVar := @ARefRecord;
        if pUndoRefVar <> nil then
            UndoRefValue := pUndoRefVar^;
        if AFreeRec <> nil then
            UndoFreeRec := AFreeRec^;
    end;

    procedure DoUndo;
    begin
        if pUndoVar <> nil then
            pUndoVar^ := UndoValue;
        if pUndoRefVar <> nil then
            pUndoRefVar^ := UndoRefValue;
        if UndoFreeRec <> nil then
            FreeRecord(@(UndoFreeRec.FSelfBmk));
    end;

begin
    CheckModifable;
    pRH := nil;
    pCur := PageRecordHeader(FPosition.FPageIndex, FPosition.FRecIndex);
    InitUndo;
    try
        with pCur^ do
            case FStatus of
            rsFetched, rsNewDeleted:
                begin
                    pRH := AllocateRecordInternal(rsNewUpdated);
                    pRH^.FRefRecord := pCur;
                    pRH^.FRecordNumber := pCur^.FRecordNumber;
                    SaveUndo(@FStatus, @FRefRecord, @pRH);
                    with FPosition do
                        SetRecordstatus(FPageIndex, FRecIndex, rsOld);
                    FRefRecord := pRH;
                end;
            rsOld:
                begin
                    SaveUndo(@FRefRecord.FStatus, nil, nil);
                    with FRefRecord^.FSelfBmk do
                        pRH := SetRecordstatus(FPageIndex, FRecIndex, rsNewUpdated);
                end;
            rsNewUpdated, rsNewInserted:
                pRH := pCur;
            rsEmpty, rsNewInsApplyed, rsNewUpdApplyed, rsNewDelApplyed, rsDelphiBuff:
                OCIDBError(msgOCICantModifyBuff);
            end;
        pBmk := @(pRH.FSelfBmk);
        CopyRecord(P, pBmk);
        EndUpdate(pBmk);
    except
        DoUndo;
        raise;
    end;
    SetCurrentRecord(pBmk);
end;

procedure TOCIDeadCursor.InsertRecord(P: POCIBookmark);
var
    pRH: POCIDeadRecordHeader;
    pBmk: POCIDeadBookmark;
begin
    CheckModifable;
    pRH := AllocateRecordInternal(rsNewInserted);
    pBmk := @pRH.FSelfBmk;
    try
        CopyRecord(P, pBmk);
        EndUpdate(pBmk);
    except
        FreeRecord(@pRH^.FSelfBmk);
        raise;
    end;
    SetCurrentRecord(pBmk);
    Inc(FRowsInserted);
end;

function TOCIDeadCursor.CompareBookmarks(P1, P2: POCIBookmark): sb2;
begin
    if (POCIDeadBookmark(P1).FPageIndex < POCIDeadBookmark(P2).FPageIndex) or
       (POCIDeadBookmark(P1).FPageIndex = POCIDeadBookmark(P2).FPageIndex) and
        (POCIDeadBookmark(P1).FRecIndex < POCIDeadBookmark(P2).FRecIndex) then
        Result := -1
    else
        Result := 1;
end;

function TOCIDeadCursor.BookmarkValid(P: POCIBookmark): Boolean;
begin
    try
        with POCIDeadBookmark(P)^ do
            PageRecordHeader(FPageIndex, FRecIndex);
        Result := True;
    except
        Result := False;
    end;
end;

function TOCIDeadCursor.GetNewRecord(P: POCIBookmark): POCIBookmark;
begin
    if p = nil then
        p := @FPosition;
    Result := nil;
    with POCIDeadBookmark(p)^, PageRecordHeader(FPageIndex, FRecIndex)^ do begin
        case FStatus of
             rsFetched, rsNewInserted, rsNewUpdated, rsNewDeleted,
             rsNewInsApplyed, rsNewUpdApplyed, rsNewDelApplyed:
                Result := @FSelfBmk;
             rsOld:
                Result := @(FRefRecord^.FSelfBmk);
        end;
    end;
end;

function TOCIDeadCursor.GetOldRecord(P: POCIBookmark): POCIBookmark;
begin
    if p = nil then
        p := @FPosition;
    Result := nil;
    with POCIDeadBookmark(p)^, PageRecordHeader(FPageIndex, FRecIndex)^ do begin
        case FStatus of
             rsFetched, rsNewInserted, rsNewUpdated, rsNewDeleted,
             rsNewInsApplyed, rsNewUpdApplyed, rsNewDelApplyed:
                Result := @(FRefRecord^.FSelfBmk);
             rsOld:
                Result := @FSelfBmk;
        end;
    end;
end;

procedure TOCIDeadCursor.CheckModifable;
begin
    if InApplyUpdates then
        OCIDBError(msgOCIRecurseApply);
    if not Assigned(OnUpdateRecord) then
        OCIDBError(msgOCINoApplyCallback);
end;

procedure TOCIDeadCursor.ApplyUpdates(P: POCIBookmark);
var
    prevStatusFilter: TOCIRecordStatuses;
    updAction: TOCIUpdateAction;
    oldStatus, newStatus: TOCIRecordStatus;
begin
    CheckModifable;
    prevStatusFilter := StatusFilter;
    FInApplyUpdates := True;
    try
        StatusFilter := [rsNewInserted, rsNewUpdated, rsNewDeleted];
        if P = nil then
            First
        else
            Bookmark := P;
        while not EOF do begin
            updAction := uaFail;
            repeat
                try
                    OnUpdateRecord(Self, updAction);
                except on E: Exception do
                    if Assigned(OnUpdateError) then
                        OnUpdateError(Self, E, updAction)
                    else begin
                        Application.HandleException(Self);
                        updAction := uaAbort;
                    end;
                end;
            until updAction <> uaRetry;
            case updAction of
                uaAbort:
                    Break;
                uaFail:
                    OCIDBError(msgOCIApplyFailed);
                uaApplied:
                    begin
                        with FPosition do
                            oldStatus := PageRecordHeader(FPageIndex, FRecIndex)^.FStatus;
                        case oldStatus of
                            rsNewInserted: newStatus := rsNewInsApplyed;
                            rsNewUpdated: newStatus := rsNewUpdApplyed;
                            rsNewDeleted: newStatus := rsNewDelApplyed;
                        end;
                        with FPosition do
                            SetRecordStatus(FPageIndex, FRecIndex, newStatus);
                    end;
            end;
            if P = nil then
                Skip(1)
            else
                Break;
        end;
    finally
        StatusFilter := prevStatusFilter;
        FInApplyUpdates := False;
    end;
end;

{
procedure TOCIDeadCursor.CancelUpdates(P: POCIBookmark);
begin
end;

procedure TOCIDeadCursor.CommitUpdates(P: POCIBookmark);
begin
end;
}

end.

